/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with {library}; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package ch.hortis.sonar.service;

import ch.hortis.sonar.model.MavenProject;
import ch.hortis.sonar.model.Metric;
import ch.hortis.sonar.model.Snapshot;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;
import org.sonar.commons.database.DatabaseManager;
import org.sonar.commons.measures.Measure;

import javax.persistence.Query;
import java.util.*;

public class MeasuresDao {

  private DatabaseManager databaseManager;
  private Map<String, Metric> metricsByName = new HashMap<String, Metric>();

  public MeasuresDao(DatabaseManager databaseManager) {
    this.databaseManager = databaseManager;
  }

  public Metric getMetric(Metric metric) {
    return getMetricsByName().get(metric.getKey());
  }

  public Metric getMetric(String metricName) {
    return getMetricsByName().get(metricName);
  }

  public Collection<Metric> getMetrics() {
    return getMetricsByName().values();
  }

  public Collection<Metric> getEnabledMetrics() {
    return CollectionUtils.select(getMetricsByName().values(), new Predicate(){
      public boolean evaluate(Object o) {
        return ((Metric)o).getEnabled();
      }
    });
  }

  public void disableAutomaticMetrics() {
    databaseManager.createQuery("update " + Metric.class.getSimpleName() + " m set m.enabled=false where m.userManaged=false").executeUpdate();
    databaseManager.commit();
    metricsByName.clear();
  }

  public void persistMetric(Metric metric) {
    Metric dbMetric = getMetric(metric);
    if (dbMetric!=null) {
      dbMetric.setComparable(metric.isComparable());
      dbMetric.setDescription(metric.getDescription());
      dbMetric.setDirection(metric.getDirection());
      dbMetric.setDomain(metric.getDomain());
      dbMetric.setEnabled(metric.getEnabled());
      dbMetric.setQualitative(metric.getQualitative());
      dbMetric.setName(metric.getName());
      dbMetric.setType(metric.getType());
      dbMetric.setUserManaged(metric.getUserManaged());
      dbMetric.setOrigin(metric.getOrigin());
      databaseManager.getEntityManager().merge(dbMetric);

    } else {
      databaseManager.getEntityManager().persist(metric);
    }

    metricsByName.clear();
  }

  private Map<String, Metric> getMetricsByName() {
    if (metricsByName.isEmpty()) {
      List<Metric> metrics = databaseManager.createQuery("SELECT m from " + Metric.class.getSimpleName() + " m").getResultList();
      for (Metric metric : metrics) {
        metricsByName.put(metric.getKey(), metric);
      }
    }
    return metricsByName;
  }

  public List<Measure> getMeasures(Integer snapshotId) {
    Query query = databaseManager.createQuery(
      "SELECT m from " + Measure.class.getSimpleName() + " m where m.snapshotId=:snapshotId");
    query.setParameter("snapshotId", snapshotId);
    return query.getResultList();
  }

  public Snapshot getSnapshot(int snapshotId) {
    Query query = databaseManager.createQuery(
      "SELECT s FROM " + Snapshot.class.getSimpleName() + " s WHERE s.id=:id");
    query.setParameter("id", snapshotId);
    return (Snapshot) query.getSingleResult();
  }

  public String getSnapshotProjectName(int snapshotId) {
    Query query = databaseManager.createQuery(
      "SELECT p.name FROM " + MavenProject.class.getSimpleName() + " p, Snapshot s WHERE s.mavenProject=p AND s.id=:id");
    query.setParameter("id", snapshotId);
    return (String) query.getSingleResult();
  }

  public List<Snapshot> getChildSnapshots(Snapshot snapshot, boolean includeSubmodules) {
    List<Snapshot> result = new ArrayList<Snapshot>();
    Query query = databaseManager.createQuery(
      "SELECT s FROM " + Snapshot.class.getSimpleName() + " s WHERE s.parent = :parent");
    query.setParameter("parent", snapshot);
    List<Snapshot> childrens = query.getResultList();
    result.addAll(childrens);
    if (includeSubmodules) {
      for (Snapshot child : childrens) {
        List<Snapshot> subchildrens = getChildSnapshots(child, true);
        result.addAll(subchildrens);
      }
    }
    return result;
  }
}
